from dataclasses import asdict, dataclass
from typing import Any

import pytest
from tests.utils import random_string

from scyllapy import Scylla, extra_types
from scyllapy.exceptions import ScyllaPyDBError


@pytest.mark.anyio
@pytest.mark.parametrize(
    ("type_cls", "type_name", "test_val"),
    [
        (extra_types.TinyInt, "TINYINT", 1),
        (extra_types.SmallInt, "SMALLINT", 1),
        (extra_types.BigInt, "BIGINT", 1),
        (extra_types.Double, "DOUBLE", 1.0),
    ],
)
async def test_int_types(
    scylla: Scylla,
    type_cls: Any,
    type_name: str,
    test_val: Any,
) -> None:
    table_name = random_string(4)

    await scylla.execute(
        f"CREATE TABLE {table_name} (id {type_name}, PRIMARY KEY (id))",
    )
    insert_query = f"INSERT INTO {table_name}(id) VALUES (?)"
    with pytest.raises(ScyllaPyDBError):
        await scylla.execute(insert_query, [test_val])

    await scylla.execute(insert_query, [type_cls(test_val)])

    result = await scylla.execute(f"SELECT * FROM {table_name}")
    rows = result.all()
    assert len(rows) == 1
    assert rows[0] == {"id": test_val}


@pytest.mark.anyio
async def test_counter(scylla: Scylla) -> None:
    table_name = random_string(4)
    await scylla.execute(
        f"CREATE TABLE {table_name} (id INT, count COUNTER, PRIMARY KEY (id))",
    )

    query = f"UPDATE {table_name} SET count = count + ? WHERE id = ?"

    with pytest.raises(ScyllaPyDBError):
        await scylla.execute(query, [1, 1])

    await scylla.execute(query, [extra_types.Counter(1), 1])

    res = await scylla.execute(f"SELECT * FROM {table_name}")
    rows = res.all()
    assert len(rows) == 1
    assert rows[0] == {"id": 1, "count": 1}


@pytest.mark.anyio
async def test_unset(scylla: Scylla) -> None:
    table_name = random_string(4)
    await scylla.execute(f"CREATE TABLE {table_name} (id INT PRIMARY KEY, name TEXT)")

    await scylla.execute(
        f"INSERT INTO {table_name}(id, name) VALUES (?, ?)",
        [1, extra_types.Unset()],
    )


@pytest.mark.anyio
async def test_udts(scylla: Scylla) -> None:
    @dataclass
    class TestUDT(extra_types.ScyllaPyUDT):
        id: int
        name: str

    table_name = random_string(4)

    udt_val = TestUDT(id=1, name="test")
    await scylla.execute(f"CREATE TYPE test_udt{table_name} (id int, name text)")
    await scylla.execute(
        f"CREATE TABLE {table_name} "
        f"(id INT PRIMARY KEY, udt_col frozen<test_udt{table_name}>)",
    )
    await scylla.execute(
        f"INSERT INTO {table_name} (id, udt_col) VALUES (?, ?)",
        [1, udt_val],
    )

    res = await scylla.execute(f"SELECT * FROM {table_name}")
    assert res.all() == [{"id": 1, "udt_col": asdict(udt_val)}]


@pytest.mark.anyio
async def test_nested_udts(scylla: Scylla) -> None:
    @dataclass
    class NestedUDT(extra_types.ScyllaPyUDT):
        one: int
        two: str

    @dataclass
    class TestUDT(extra_types.ScyllaPyUDT):
        id: int
        name: str
        nested: NestedUDT

    table_name = random_string(4)

    udt_val = TestUDT(id=1, name="test", nested=NestedUDT(one=1, two="2"))
    await scylla.execute(f"CREATE TYPE nested_udt{table_name} (one int, two text)")
    await scylla.execute(
        f"CREATE TYPE test_udt{table_name} "
        f"(id int, name text, nested frozen<nested_udt{table_name}>)",
    )
    await scylla.execute(
        f"CREATE TABLE {table_name} "
        f"(id INT PRIMARY KEY, udt_col frozen<test_udt{table_name}>)",
    )
    await scylla.execute(
        f"INSERT INTO {table_name} (id, udt_col) VALUES (?, ?)",
        [1, udt_val],
    )

    res = await scylla.execute(f"SELECT * FROM {table_name}")
    assert res.all() == [{"id": 1, "udt_col": asdict(udt_val)}]


@pytest.mark.parametrize(
    ["typ", "val"],
    [
        ("BIGINT", 1),
        ("TINYINT", 1),
        ("SMALLINT", 1),
        ("INT", 1),
        ("FLOAT", 1.0),
        ("DOUBLE", 1.0),
    ],
)
@pytest.mark.anyio
async def test_autocast_positional(scylla: Scylla, typ: str, val: Any) -> None:
    table_name = random_string(4)
    await scylla.execute(f"CREATE TABLE {table_name}(id INT PRIMARY KEY, val {typ})")
    prepared = await scylla.prepare(f"INSERT INTO {table_name}(id, val) VALUES (?, ?)")
    await scylla.execute(prepared, [1, val])



import pytest
from tests.utils import random_string

from scyllapy import Scylla


@pytest.mark.anyio
async def test_udt_parsing(scylla: Scylla) -> None:
    table_name = random_string(4)
    await scylla.execute(f"CREATE TYPE test_udt{table_name} (id int, name text)")
    await scylla.execute(
        f"CREATE TABLE {table_name} "
        f"(id int PRIMARY KEY, udt_col frozen<test_udt{table_name}>)",
    )
    await scylla.execute(
        f"INSERT INTO {table_name} (id, udt_col) VALUES (1, {{id: 1, name: 'test'}})",
    )
    res = await scylla.execute(f"SELECT * FROM {table_name}")
    assert res.all() == [{"id": 1, "udt_col": {"id": 1, "name": "test"}}]



import dataclasses
from typing import Any, List

from ._internal.extra_types import BigInt, Counter, Double, SmallInt, TinyInt, Unset

try:
    import pydantic
except ImportError:
    pydantic = None


class ScyllaPyUDT:
    """
    Class for declaring UDT models.

    This class is a mixin for models like dataclasses and pydantic models,
    or classes that have `__slots__` attribute.

    It can be further extended to support other model types.
    """

    def __dump_udt__(self) -> List[Any]:
        """
        Method to dump UDT models to a dict.

        This method returns a list of values in the order of the UDT fields.
        Because in the protocol, UDT fields should be sent in the same order as
        they were declared.
        """
        if dataclasses.is_dataclass(self):
            values = []
            for field in dataclasses.fields(self):
                values.append(getattr(self, field.name))
            return values
        if pydantic is not None and isinstance(self, pydantic.BaseModel):
            values = []
            for param in self.__class__.__signature__.parameters:
                values.append(getattr(self, param))
            return values
        if hasattr(self, "__slots__"):
            values = []
            for slot in self.__slots__:
                values.append(getattr(self, slot))
            return values
        raise ValueError("Unsupported model type")


__all__ = ("BigInt", "Counter", "Double", "SmallInt", "TinyInt", "Unset", "ScyllaPyUDT")



from dataclasses import dataclass

import pytest
from tests.utils import random_string

from scyllapy import Scylla


@pytest.mark.anyio
async def test_empty_scalars(scylla: Scylla) -> None:
    table_name = random_string(4)
    await scylla.execute(f"CREATE TABLE {table_name} (id INT PRIMARY KEY)")
    res = await scylla.execute(f"SELECT id FROM {table_name}")

    assert res.all() == []
    assert res.scalars() == []


@pytest.mark.anyio
async def test_as_class(scylla: Scylla) -> None:
    @dataclass
    class TestDTO:
        id: int

    table_name = random_string(4)
    await scylla.execute(f"CREATE TABLE {table_name} (id INT PRIMARY KEY)")
    await scylla.execute(f"INSERT INTO {table_name}(id) VALUES (?)", [42])
    res = await scylla.execute(f"SELECT id FROM {table_name}")

    assert res.all(as_class=TestDTO) == [TestDTO(id=42)]


@pytest.mark.anyio
async def test_udt_as_dataclass(scylla: Scylla) -> None:
    @dataclass
    class UDTType:
        id: int
        name: str

    @dataclass
    class TestDTO:
        id: int
        udt_col: UDTType

        def __post_init__(self) -> None:
            if not isinstance(self.udt_col, UDTType):
                self.udt_col = UDTType(**self.udt_col)

    table_name = random_string(4)
    await scylla.execute(f"CREATE TYPE test_udt{table_name} (id int, name text)")
    await scylla.execute(
        f"CREATE TABLE {table_name} "
        f"(id int PRIMARY KEY, udt_col frozen<test_udt{table_name}>)",
    )
    await scylla.execute(
        f"INSERT INTO {table_name} (id, udt_col) VALUES (1, {{id: 1, name: 'test'}})",
    )
    res = await scylla.execute(f"SELECT * FROM {table_name}")
    assert res.all(as_class=TestDTO) == [
        TestDTO(id=1, udt_col=UDTType(id=1, name="test")),
    ]



class LatencyAwareness:
    def __init__(
        self,
        *,
        minimum_measurements: int | None = None,
        retry_period: int | None = None,
        exclusion_threshold: float | None = None,
        update_rate: int | None = None,
        scale: int | None = None,
    ) -> None: ...
    """
    Build latency awareness for balancing policy.

    :param minimum_measurements: Minimum number of measurements to consider a host
        as eligible for query plans.
    :param retry_period: Number of milliseconds to wait before attempting to refresh
        the latency information of a host.
    :param exclusion_threshold: Maximum ratio of hosts to exclude from query plans.
        For example, if set to 2, the resulting policy excludes nodes that are
        more than twice slower than the fastest node.
    :param update_rate: Number of milliseconds between measurements.
    :param scale: provides control on how the weight given to older latencies decreases
        over time.
    """

class LoadBalancingPolicy:
    """
    Load balancing policy.

    Useful to control how the driver distributes queries among nodes.
    Can be applied to profiles.
    """

    @classmethod
    async def build(
        cls,
        *,
        token_aware: bool | None = None,
        prefer_rack: str | None = None,
        prefer_datacenter: str | None = None,
        permit_dc_failover: bool | None = None,
        shuffling_replicas: bool | None = None,
    ) -> LoadBalancingPolicy: ...
    """
    Construct load balancing policy.

    It requires to be async, becausse it needs to start a background task.

    :param token_aware: Whether to use token aware routing.
    :param prefer_rack: Name of the rack to prefer.
    :param prefer_datacenter: Name of the datacenter to prefer.
    :param permit_dc_failover: Whether to allow datacenter failover.
    :param shuffling_replicas: Whether to shuffle replicas.
    """

    async def with_latency_awareness(
        self,
        latency_awareness: LatencyAwareness,
    ) -> None: ...



from dataclasses import dataclass

import pytest
from tests.utils import random_string

from scyllapy import Scylla
from scyllapy.query_builder import Select


@pytest.mark.anyio
async def test_scalars(scylla: Scylla) -> None:
    table_name = random_string(4)
    await scylla.execute(
        f"CREATE TABLE IF NOT EXISTS {table_name} (id INT PRIMARY KEY)",
    )
    vals = list(range(10))
    for i in vals:
        await scylla.execute(f"INSERT INTO {table_name}(id) VALUES (?)", [i])
    res = await scylla.execute(f"SELECT id FROM {table_name}", paged=True)
    async for col in res.scalars():
        assert col in vals


@pytest.mark.anyio
async def test_dicts(scylla: Scylla) -> None:
    table_name = random_string(4)
    await scylla.execute(
        f"CREATE TABLE IF NOT EXISTS {table_name} (id INT PRIMARY KEY, val INT)",
    )
    vals = list(range(10))
    for i in vals:
        await scylla.execute(
            f"INSERT INTO {table_name}(id, val) VALUES (?, ?)",
            [i, -i],
        )
    res = await scylla.execute(f"SELECT id, val FROM {table_name}", paged=True)
    async for row in res:
        assert row["id"] in vals
        assert row["val"] == -row["id"]


@pytest.mark.anyio
async def test_dtos(scylla: Scylla) -> None:
    @dataclass
    class TestDTO:
        id: int
        val: int

    table_name = random_string(4)
    await scylla.execute(
        f"CREATE TABLE IF NOT EXISTS {table_name} (id INT PRIMARY KEY, val INT)",
    )
    vals = list(range(10))
    for i in vals:
        await scylla.execute(
            f"INSERT INTO {table_name}(id, val) VALUES (?, ?)",
            [i, -i],
        )
    res = await scylla.execute(f"SELECT id, val FROM {table_name}", paged=True)
    async for row in res.as_cls(TestDTO):
        assert row.id in vals
        assert row.val == -row.id


@pytest.mark.anyio
async def test_paged_select_qb(scylla: Scylla) -> None:
    table_name = random_string(4)
    await scylla.execute(
        f"CREATE TABLE IF NOT EXISTS {table_name} (id INT PRIMARY KEY, val INT)",
    )
    vals = list(range(10))
    for i in vals:
        await scylla.execute(
            f"INSERT INTO {table_name}(id, val) VALUES (?, ?)",
            [i, -i],
        )
    res = await Select(table_name).execute(scylla, paged=True)
    async for row in res:
        assert row["id"] in vals
        assert row["val"] == -row["id"]



import pytest
from tests.utils import random_string

from scyllapy import Consistency, ExecutionProfile, Query, Scylla
from scyllapy.exceptions import ScyllaPyDBError


@pytest.mark.anyio
async def test_wrong_consistency(scylla: Scylla) -> None:
    profile = ExecutionProfile(consistency=Consistency.ANY)
    table_name = random_string(4)
    await scylla.execute(f"CREATE TABLE {table_name}(id INT PRIMARY KEY)")
    query = Query(f"SELECT * FROM {table_name} WHERE id = ?", profile=profile)
    with pytest.raises(ScyllaPyDBError, match=".*only supported for writes.*"):
        await scylla.execute(query, [1])

    await scylla.execute(query.with_profile(None), [1])
